from rejson import Path
from functools import reduce
from typing import List, Dict, Iterable
import json
import logging

logger = logging.getLogger("reem.datatypes")


def append_to_path(existing, addition):
    if existing == Path.rootPath():
        return Path.rootPath() + addition
    return "{}.{}".format(existing, addition)


def bytes_to_dict(by):
    logger.debug("Received: {}".format(by))
    return json.loads(by.decode("utf-8"))


def path_to_key_sequence(path: str):
    if path == Path.rootPath():
        return []
    return path.split(".")[1:]


def key_sequence_to_path(sequence: List[str]):
    return Path.rootPath() + ".".join(sequence)


def copy_dictionary_without_paths(dictionary: Dict, key_sequences: List[List[str]]):
    """
    Return a copy of dictionary with the paths formed by key_sequences removed
    :param key_sequences: 2D List of Strings where each internal list is a sequence of key accesses from dictionary root
    :param dictionary: The dictionary to copy
    :return: dict
    """
    ret = {}
    possibles = [ks for ks in key_sequences if len(ks) == 1]
    possibles = set(reduce(lambda x, y: x + y, possibles, []))
    for k, v in dictionary.items():
        if k in possibles:
            continue
        if type(v) == dict:
            ret[k] = copy_dictionary_without_paths(v, [ks[1:] for ks in key_sequences if len(ks) > 1])
        else:
            ret[k] = v
    return ret


def extract_object(dictionary: Dict, key_sequence: List[str]):
    ret = dictionary
    for k in key_sequence:
        ret = ret[k]
    return ret


def filter_paths_by_prefix(all_paths: Iterable[str], prefix: str):
    return [path for path in all_paths if path[:len(prefix)] == prefix]


def insert_into_dictionary(dictionary: Dict, path: str, value):
    key_sequence = path_to_key_sequence(path)
    parent = dictionary
    for key in key_sequence[:-1]:
        if key not in parent.keys():
            parent[key] = {}
        parent = parent[key]
    parent[key_sequence[-1]] = value


def get_special_paths(set_path: str, set_value, sp_to_label: Dict, label_to_ship: Dict):
    """
    Get info on how to update sp_to_label according to the user uploading `set_value` in location `set_path`
    :param set_path: ".key1.key0" kind of string indicating path within parent that set_value is supposed to be set
    :param set_value: anything to be inserted into data. Needs to be able to be sent to redis or a dict
    :param sp_to_label: dictionary we are going to be updating
    :param label_to_ship: dictionary giving the ship corresponding to a certain label
    :return: set( tuple(str:paths added, str:labels added) ), set( str:paths removed )
    """
    additions = set()
    if type(set_value) is not dict:
        if set_path in sp_to_label:
            if label_to_ship[sp_to_label[set_path]].check_fit(set_value):
                return additions

        # If this is something a ship covers, add it as a special path
        for ship in label_to_ship.values():
            if ship.check_fit(set_value):
                additions.add( (set_path, ship.get_label()) )

    else:
        for k, v in set_value.items():
            assert check_valid_key_name(k), "Invalid Key: {}".format(k)
            child_path = "{}.{}".format(set_path, k)
            child_add = get_special_paths(child_path, v, sp_to_label, label_to_ship)
            additions = additions.union(child_add)

    return additions


def check_valid_key_name(name):
    if type(name) not in [str]:
        return False
    bad_chars = ["*", ".", "&&&&"]
    for k in bad_chars:
        if k in name:
            return False
    return True


SEPARATOR_CHARACTER = "&&&&"
ROOT_VALUE_SEQUENCE = "%%%%"


from threading import Thread
import rejson
from .helper_functions import *


class MetadataListener:
    def __init__(self, interface):
        self.client = rejson.Client(host=interface.hostname)
        self.pubsub = self.client.pubsub()
        self.pubsub.psubscribe(['__keyspace@0__:*'])
        self.listeners = {}
        super().__init__()

    def add_listener(self, key_name, reader):
        listen_name = "__keyspace@0__:{}".format(key_name)
        if listen_name not in self.listeners:
            self.listeners[listen_name] = []
        self.listeners[listen_name].append(reader)

    def flush(self):
        while True:
            item = self.pubsub.get_message()
            if item is None:
                break
            channel = item['channel'].decode("utf_8")
            if channel in self.listeners:
                for listener in self.listeners[channel]:
                    listener.pull_metadata = True


class PathHandler:
    def __init__(self, writer, reader, initial_path):
        self.writer = writer
        self.reader = reader
        self.path = initial_path

    def __getitem__(self, item):
        assert check_valid_key_name(item)
        self.path = append_to_path(self.path, item)
        return self

    def __setitem__(self, instance, value):
        assert check_valid_key_name(instance)
        self.path = append_to_path(self.path, instance)
        self.writer.send_to_redis(self.path, value)


class ReadablePathHandler(PathHandler):
    def read(self):
        server_value = self.reader.read_from_redis(self.path)
        root_value_read_name = "{}ROOT{}".format(ROOT_VALUE_SEQUENCE, ROOT_VALUE_SEQUENCE)
        if type(server_value)==dict and root_value_read_name in server_value:
            return server_value[root_value_read_name]
        return server_value


class ActiveSubscriberPathHandler(PathHandler):
    def read(self):
        return_val = self.reader.local_copy
        dissect_path = self.path[1:]
        if len(dissect_path) == 0:
            pass
        elif "." in dissect_path:
            for key in dissect_path.split("."):
                return_val = return_val[key]
        else:
            return_val = return_val[dissect_path]
        if type(return_val) == dict:
            return copy_dictionary_without_paths(return_val, [])
        return return_val

    def __setitem__(self, instance, value):
        raise Exception("Cannot set value for a subscriber")


class ChannelListener(Thread):
    def __init__(self, interface, channel_name, callback_function, kwargs):
        self.client = rejson.Client(host=interface.hostname)
        self.pubsub = self.client.pubsub()
        self.pubsub.psubscribe([channel_name])
        self.callback_function = callback_function
        self.kwargs = kwargs
        self.first_item_seen = False
        super().__init__()

    def run(self):
        for item in self.pubsub.listen():
            # First Item is a generic message that we need to get rid of
            if not self.first_item_seen:
                self.first_item_seen = True
                continue
            channel = item['channel'].decode("utf_8")
            message = item['data']
            self.callback_function(channel=channel, message=message, **self.kwargs)

"""
Naming:
Rapid Extendable Middleware
Redis Medium Extendable Middleware
Redis Robotic Communication

Redis Extendable Middleware

Extendable Efficient Redis Middleware
Redis Extendable Efficient Middleware
reem
"""